"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AptosContext = exports.APTOS_COIN = void 0;
const ethers_1 = require("ethers");
const js_sha3_1 = require("js-sha3");
const utils_1 = require("ethers/lib/utils");
const aptos_1 = require("aptos");
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
const contracts_1 = require("./contracts");
const utils_2 = require("./utils");
exports.APTOS_COIN = '0x1::aptos_coin::AptosCoin';
/**
 * @category Aptos
 */
class AptosContext extends connect_sdk_1.TokenBridgeAbstract {
    constructor(wormholeInstance) {
        super();
        this.type = connect_sdk_1.Context.APTOS;
        this.wormhole = wormholeInstance;
        const rpc = this.wormhole.conf.rpcs.aptos;
        if (rpc === undefined)
            throw new Error('No Aptos rpc configured');
        this.aptosClient = new aptos_1.AptosClient(rpc);
        this.coinClient = new aptos_1.CoinClient(this.aptosClient);
        this.contracts = new contracts_1.AptosContracts(this.wormhole, this.aptosClient);
    }
    async startTransfer(token, amount, sendingChain, senderAddress, recipientChain, recipientAddress, relayerFee = '0') {
        const destContext = this.wormhole.getContext(recipientChain);
        const recipientChainId = this.wormhole.toChainId(recipientChain);
        let recipientAccount = recipientAddress;
        // get token account for solana
        if (recipientChainId === connect_sdk_1.MAINNET_CHAINS.solana) {
            let tokenId = token;
            if (token === connect_sdk_1.NATIVE) {
                tokenId = {
                    address: exports.APTOS_COIN,
                    chain: 'aptos',
                };
            }
            recipientAccount = await this.wormhole.getSolanaRecipientAddress(recipientChain, tokenId, recipientAddress);
        }
        const formattedRecipientAccount = (0, utils_1.arrayify)(destContext.formatAddress(recipientAccount));
        let coinType;
        if (token === connect_sdk_1.NATIVE) {
            coinType = exports.APTOS_COIN;
        }
        else {
            coinType = await this.mustGetForeignAsset(token, sendingChain);
        }
        const payload = (0, utils_2.transferFromAptos)(this.contracts.mustGetBridge(sendingChain), coinType, amount.toString(), recipientChainId, formattedRecipientAccount, relayerFee, undefined);
        return payload;
    }
    async startTransferWithPayload(token, amount, sendingChain, senderAddress, recipientChain, recipientAddress, payload) {
        throw new Error('Aptos send with payload not implemented');
    }
    formatAddress(address) {
        return (0, utils_1.arrayify)((0, utils_1.zeroPad)(address, 32));
    }
    parseAddress(address) {
        return (0, utils_1.hexlify)((0, utils_1.stripZeros)(address));
    }
    async formatAssetAddress(address) {
        if (!(0, utils_2.isValidAptosType)(address)) {
            throw new Error(`Unable to format Aptos asset address: ${address}`);
        }
        return (0, connect_sdk_1.hexToUint8Array)((0, js_sha3_1.sha3_256)(address));
    }
    async parseAssetAddress(address) {
        const bridge = this.contracts.mustGetBridge('aptos');
        const assetType = await (0, utils_2.getTypeFromExternalAddress)(this.aptosClient, bridge, address);
        if (!assetType)
            throw new Error(`Unable to parse Aptos asset address: ${address}`);
        return assetType;
    }
    async getForeignAsset(tokenId, chain) {
        const chainId = this.wormhole.toChainId(tokenId.chain);
        const toChainId = this.wormhole.toChainId(chain);
        if (toChainId === chainId)
            return tokenId.address;
        const { token_bridge } = this.wormhole.mustGetContracts(chain);
        if (!token_bridge)
            throw new Error('token bridge contract not found');
        const tokenContext = this.wormhole.getContext(tokenId.chain);
        const formattedAddr = await tokenContext.formatAssetAddress(tokenId.address);
        return await (0, utils_2.getForeignAssetAptos)(this.aptosClient, token_bridge, chainId, (0, utils_1.hexlify)(formattedAddr));
    }
    async mustGetForeignAsset(tokenId, chain) {
        const addr = await this.getForeignAsset(tokenId, chain);
        if (!addr)
            throw new Error('token not registered');
        return addr;
    }
    async fetchTokenDecimals(tokenAddr, chain) {
        const coinType = `0x1::coin::CoinInfo<${tokenAddr}>`;
        const decimals = (await this.aptosClient.getAccountResource(tokenAddr.split('::')[0], coinType)).data.decimals;
        return decimals;
    }
    async parseMessageFromTx(tx, chain) {
        const transaction = await this.aptosClient.getTransactionByHash(tx);
        if (transaction.type !== 'user_transaction') {
            throw new Error(`${tx} is not a user_transaction`);
        }
        const userTransaction = transaction;
        const message = userTransaction.events.find((event) => event.type.endsWith('WormholeMessage'));
        if (!message || !message.data) {
            throw new Error(`WormholeMessage not found for ${tx}`);
        }
        const { payload, sender, sequence } = message.data;
        const parsed = (0, connect_sdk_1.parseTokenTransferPayload)(Buffer.from(payload.slice(2), 'hex'));
        const tokenContext = this.wormhole.getContext(parsed.tokenChain);
        const destContext = this.wormhole.getContext(parsed.toChain);
        const tokenAddress = await tokenContext.parseAssetAddress((0, utils_1.hexlify)(parsed.tokenAddress));
        const tokenChain = this.wormhole.toChainName(parsed.tokenChain);
        // make sender address even-length
        const emitter = (0, utils_1.hexlify)(sender, {
            allowMissingPrefix: true,
            hexPad: 'left',
        });
        const parsedMessage = {
            sendTx: tx,
            sender: userTransaction.sender,
            amount: ethers_1.BigNumber.from(parsed.amount),
            payloadID: Number(parsed.payloadType),
            recipient: destContext.parseAddress((0, utils_1.hexlify)(parsed.to)),
            toChain: this.wormhole.toChainName(parsed.toChain),
            fromChain: this.wormhole.toChainName(chain),
            tokenAddress,
            tokenChain,
            tokenId: {
                chain: tokenChain,
                address: tokenAddress,
            },
            sequence: ethers_1.BigNumber.from(sequence),
            emitterAddress: (0, utils_1.hexlify)(this.formatAddress(emitter)),
            block: Number(userTransaction.version),
            gasFee: ethers_1.BigNumber.from(userTransaction.gas_used).mul(userTransaction.gas_unit_price),
        };
        return [parsedMessage];
    }
    async getNativeBalance(walletAddress, chain) {
        return await this.checkBalance(walletAddress, exports.APTOS_COIN);
    }
    async getTokenBalance(walletAddress, tokenId, chain) {
        const address = await this.getForeignAsset(tokenId, chain);
        if (!address)
            return null;
        const balance = await this.checkBalance(walletAddress, address);
        return balance ? ethers_1.BigNumber.from(balance) : null;
    }
    async checkBalance(walletAddress, coinType) {
        try {
            const balance = await this.coinClient.checkBalance(walletAddress, {
                coinType,
            });
            return ethers_1.BigNumber.from(balance);
        }
        catch (e) {
            if ((e instanceof aptos_1.Types.ApiError || e.errorCode === 'resource_not_found') &&
                e.status === 404) {
                return ethers_1.BigNumber.from(0);
            }
            throw e;
        }
    }
    async completeTransfer(destChain, signedVAA, overrides, payerAddr) {
        const payload = await (0, utils_2.redeemOnAptos)(this.aptosClient, this.contracts.mustGetBridge(destChain), signedVAA);
        return payload;
    }
    async isTransferCompleted(destChain, signedVaa) {
        return await (0, utils_2.getIsTransferCompletedAptos)(this.aptosClient, this.contracts.mustGetBridge(destChain), (0, utils_1.arrayify)(signedVaa));
    }
    getTxIdFromReceipt(hash) {
        return hash;
    }
    parseRelayerPayload(payload) {
        throw new Error('relaying is not supported on aptos');
    }
    async getCurrentBlock() {
        throw new Error('Aptos getCurrentBlock not implemented');
    }
}
exports.AptosContext = AptosContext;
//# sourceMappingURL=context.js.map