import { AptosClient, TokenTypes } from 'aptos';
import { ChainId } from '@wormhole-foundation/connect-sdk';
/**
 * Test if given string is a valid fully qualified type of moduleAddress::moduleName::structName.
 * @param str String to test
 * @returns Whether or not given string is a valid type
 */
export declare const isValidAptosType: (str: string) => boolean;
/**
 * Given a hash, returns the fully qualified type by querying the corresponding TypeInfo.
 * @param client Client used to transfer data to/from Aptos node
 * @param tokenBridgeAddress Address of token bridge
 * @param fullyQualifiedTypeHash Hash of fully qualified type
 * @returns The fully qualified type associated with the given hash
 */
export declare function getTypeFromExternalAddress(client: AptosClient, tokenBridgeAddress: string, fullyQualifiedTypeHash: string): Promise<string | null>;
/**
 * Get creator address, collection name, token name, and property version from
 * a token hash. Note that this method is meant to be used for native tokens
 * that have already been registered in the NFT bridge.
 *
 * The token hash is stored in the `tokenId` field of NFT transfer VAAs and
 * is calculated by the operations in `deriveTokenHashFromTokenId`.
 * @param client
 * @param nftBridgeAddress
 * @param tokenHash Token hash
 * @returns Token ID
 */
export declare const getTokenIdFromTokenHash: (client: AptosClient, nftBridgeAddress: string, tokenHash: Uint8Array) => Promise<TokenTypes.TokenId>;
/**
 * Returns module address from given fully qualified type/module address.
 * @param str FQT or module address
 * @returns Module address
 */
export declare const coalesceModuleAddress: (str: string) => string;
/**
 * The NFT bridge creates resource accounts, which in turn create a collection
 * and mint a single token for each transferred NFT. This method derives the
 * address of that resource account from the given origin chain and address.
 * @param nftBridgeAddress
 * @param originChain
 * @param originAddress External address of NFT on origin chain
 * @returns Address of resource account
 */
export declare const deriveResourceAccountAddress: (nftBridgeAddress: string, originChainId: ChainId, originAddress: Uint8Array) => Promise<string | null>;
/**
 * Derive the module address for an asset defined by the given origin chain and address.
 * @param tokenBridgeAddress Address of token bridge (32 bytes)
 * @param originChain Chain ID of chain that original asset is from
 * @param originAddress Native address of asset
 * @returns The module address for the given asset
 */
export declare const getForeignAssetAddress: (tokenBridgeAddress: string, originChain: ChainId, originAddress: string) => string | null;
/**
 * Derives the fully qualified type of the asset defined by the given origin chain and address.
 * @param tokenBridgeAddress Address of token bridge (32 bytes)
 * @param originChain Chain ID of chain that original asset is from
 * @param originAddress Native address of asset; if origin chain ID is 22 (Aptos), this is the
 * asset's fully qualified type
 * @returns The fully qualified type on Aptos for the given asset
 */
export declare const getAssetFullyQualifiedType: (tokenBridgeAddress: string, originChain: ChainId, originAddress: string) => string | null;
