"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getAssetFullyQualifiedType = exports.getForeignAssetAddress = exports.deriveResourceAccountAddress = exports.coalesceModuleAddress = exports.getTokenIdFromTokenHash = exports.getTypeFromExternalAddress = exports.isValidAptosType = void 0;
const aptos_1 = require("aptos");
const utils_1 = require("ethers/lib/utils");
const js_sha3_1 = require("js-sha3");
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
/**
 * Test if given string is a valid fully qualified type of moduleAddress::moduleName::structName.
 * @param str String to test
 * @returns Whether or not given string is a valid type
 */
const isValidAptosType = (str) => /^(0x)?[0-9a-fA-F]+::\w+::\w+$/.test(str);
exports.isValidAptosType = isValidAptosType;
/**
 * Given a hash, returns the fully qualified type by querying the corresponding TypeInfo.
 * @param client Client used to transfer data to/from Aptos node
 * @param tokenBridgeAddress Address of token bridge
 * @param fullyQualifiedTypeHash Hash of fully qualified type
 * @returns The fully qualified type associated with the given hash
 */
async function getTypeFromExternalAddress(client, tokenBridgeAddress, fullyQualifiedTypeHash) {
    // get handle
    tokenBridgeAddress = (0, connect_sdk_1.ensureHexPrefix)(tokenBridgeAddress);
    const state = (await client.getAccountResource(tokenBridgeAddress, `${tokenBridgeAddress}::state::State`)).data;
    const handle = state.native_infos.handle;
    try {
        // get type info
        const typeInfo = await client.getTableItem(handle, {
            key_type: `${tokenBridgeAddress}::token_hash::TokenHash`,
            value_type: '0x1::type_info::TypeInfo',
            key: { hash: fullyQualifiedTypeHash },
        });
        if (!typeInfo) {
            return null;
        }
        // construct type
        const moduleName = Buffer.from(typeInfo.module_name.substring(2), 'hex').toString('ascii');
        const structName = Buffer.from(typeInfo.struct_name.substring(2), 'hex').toString('ascii');
        return `${typeInfo.account_address}::${moduleName}::${structName}`;
    }
    catch {
        return null;
    }
}
exports.getTypeFromExternalAddress = getTypeFromExternalAddress;
/**
 * Get creator address, collection name, token name, and property version from
 * a token hash. Note that this method is meant to be used for native tokens
 * that have already been registered in the NFT bridge.
 *
 * The token hash is stored in the `tokenId` field of NFT transfer VAAs and
 * is calculated by the operations in `deriveTokenHashFromTokenId`.
 * @param client
 * @param nftBridgeAddress
 * @param tokenHash Token hash
 * @returns Token ID
 */
const getTokenIdFromTokenHash = async (client, nftBridgeAddress, tokenHash) => {
    const state = (await client.getAccountResource(nftBridgeAddress, `${nftBridgeAddress}::state::State`)).data;
    const handle = state.native_infos.handle;
    const { token_data_id, property_version } = (await client.getTableItem(handle, {
        key_type: `${nftBridgeAddress}::token_hash::TokenHash`,
        value_type: `0x3::token::TokenId`,
        key: {
            hash: aptos_1.HexString.fromUint8Array(tokenHash).hex(),
        },
    }));
    return { token_data_id, property_version };
};
exports.getTokenIdFromTokenHash = getTokenIdFromTokenHash;
/**
 * Returns module address from given fully qualified type/module address.
 * @param str FQT or module address
 * @returns Module address
 */
const coalesceModuleAddress = (str) => {
    return str.split('::')[0];
};
exports.coalesceModuleAddress = coalesceModuleAddress;
/**
 * The NFT bridge creates resource accounts, which in turn create a collection
 * and mint a single token for each transferred NFT. This method derives the
 * address of that resource account from the given origin chain and address.
 * @param nftBridgeAddress
 * @param originChain
 * @param originAddress External address of NFT on origin chain
 * @returns Address of resource account
 */
const deriveResourceAccountAddress = async (nftBridgeAddress, originChainId, originAddress) => {
    if (originChainId === connect_sdk_1.MAINNET_CHAINS.aptos) {
        return null;
    }
    const chainId = Buffer.alloc(2);
    chainId.writeUInt16BE(originChainId);
    const seed = Buffer.concat([chainId, Buffer.from(originAddress)]);
    const resourceAccountAddress = await aptos_1.AptosAccount.getResourceAccountAddress(nftBridgeAddress, seed);
    return resourceAccountAddress.toString();
};
exports.deriveResourceAccountAddress = deriveResourceAccountAddress;
/**
 * Derive the module address for an asset defined by the given origin chain and address.
 * @param tokenBridgeAddress Address of token bridge (32 bytes)
 * @param originChain Chain ID of chain that original asset is from
 * @param originAddress Native address of asset
 * @returns The module address for the given asset
 */
const getForeignAssetAddress = (tokenBridgeAddress, originChain, originAddress) => {
    if (originChain === connect_sdk_1.MAINNET_CHAINS.aptos) {
        return null;
    }
    // from https://github.com/aptos-labs/aptos-core/blob/25696fd266498d81d346fe86e01c330705a71465/aptos-move/framework/aptos-framework/sources/account.move#L90-L95
    const DERIVE_RESOURCE_ACCOUNT_SCHEME = Buffer.alloc(1);
    DERIVE_RESOURCE_ACCOUNT_SCHEME.writeUInt8(255);
    let chain = Buffer.alloc(2);
    chain.writeUInt16BE(originChain);
    return (0, js_sha3_1.sha3_256)(Buffer.concat([
        (0, connect_sdk_1.hex)((0, utils_1.hexZeroPad)((0, connect_sdk_1.ensureHexPrefix)(tokenBridgeAddress), 32)),
        chain,
        Buffer.from('::', 'ascii'),
        (0, connect_sdk_1.hex)((0, utils_1.hexZeroPad)((0, connect_sdk_1.ensureHexPrefix)(originAddress), 32)),
        DERIVE_RESOURCE_ACCOUNT_SCHEME,
    ]));
};
exports.getForeignAssetAddress = getForeignAssetAddress;
/**
 * Derives the fully qualified type of the asset defined by the given origin chain and address.
 * @param tokenBridgeAddress Address of token bridge (32 bytes)
 * @param originChain Chain ID of chain that original asset is from
 * @param originAddress Native address of asset; if origin chain ID is 22 (Aptos), this is the
 * asset's fully qualified type
 * @returns The fully qualified type on Aptos for the given asset
 */
const getAssetFullyQualifiedType = (tokenBridgeAddress, originChain, originAddress) => {
    // native asset
    if (originChain === connect_sdk_1.MAINNET_CHAINS.aptos) {
        // originAddress should be of form address::module::type
        if (!(0, exports.isValidAptosType)(originAddress)) {
            console.error('Invalid qualified type');
            return null;
        }
        return (0, connect_sdk_1.ensureHexPrefix)(originAddress);
    }
    // non-native asset, derive unique address
    const wrappedAssetAddress = (0, exports.getForeignAssetAddress)(tokenBridgeAddress, originChain, originAddress);
    return wrappedAssetAddress
        ? `${(0, connect_sdk_1.ensureHexPrefix)(wrappedAssetAddress)}::coin::T`
        : null;
};
exports.getAssetFullyQualifiedType = getAssetFullyQualifiedType;
//# sourceMappingURL=utils.js.map